from batou import UpdateNeeded
from batou.component import Component
import os.path


class Hello(Component):

    asdf = None

    def verify(self):
        if not os.path.exists(self.root.name):
            raise UpdateNeeded()

    def update(self):
        with open(self.root.name, 'w') as f:
            f.write(self.root.name)


class Hello1(Hello):
    pass


class Hello2(Hello):
    pass


class Hello3(Hello):
    pass


class Hello4(Hello):
    pass


class Hello5(Hello):
    pass


class Hello6(Hello):
    pass



from ..cron import CronJob, CronTab, PurgeCronTab
from batou import ConfigurationError
import batou.vfs
import os.path
import pytest


def test_collects_cronjobs_into_crontab(root):
    root.environment.vfs_sandbox = batou.vfs.Developer(root.environment, None)
    root.component += CronJob('command1', timing='* * * * *')
    root.component += CronJob('command2', timing='* * * * *')
    root.component += CronTab()
    root.component.deploy()
    crontab = open(os.path.join(
        root.environment.workdir_base, 'mycomponent/crontab')).read()
    assert 'command1' in crontab
    assert 'command2' in crontab


def test_pruge_crontab_is_empty(root):
    root.environment.vfs_sandbox = batou.vfs.Developer(root.environment, None)
    root.component += PurgeCronTab()
    root.component.deploy()
    crontab = open(os.path.join(
        root.environment.workdir_base, 'mycomponent/crontab')).read()
    assert crontab == '# Generated by batou. Do not edit.\n\n\n'


def test_empty_crontab_must_raise(root):
    root.environment.vfs_sandbox = batou.vfs.Developer(root.environment, None)
    with pytest.raises(ConfigurationError):
        root.component += CronTab()


def test_non_empty_pruge_crontab_must_raise(root):
    root.environment.vfs_sandbox = batou.vfs.Developer(root.environment, None)
    root.component += CronJob('command1', timing='* * * * *')
    root.component += CronJob('command2', timing='* * * * *')
    with pytest.raises(ConfigurationError):
        root.component += CronTab(purge=True)


def test_additional_envs_can_be_set(root):
    root.environment.vfs_sandbox = batou.vfs.Developer(root.environment, None)
    root.component += CronJob('command2', timing='* * * * *')
    root.component += CronTab(env=dict(PATH='/foo/bar'))
    root.component.deploy()
    crontab = open(os.path.join(
        root.environment.workdir_base, 'mycomponent/crontab')).read()
    assert "PATH=/foo/bar" in crontab



from batou.utils import cmd
import batou.lib.svn
import os.path
import pytest


@pytest.mark.slow
def test_runs_svn_to_clone_repository(root):
    repos_path = os.path.join(root.environment.workdir_base, 'repos')
    cmd('svnadmin create ' + repos_path)
    cmd('svn checkout file://{dir} upstream; cd upstream;'
        'touch foo; svn add foo; svn commit -m "bar"'.format(dir=repos_path))
    root.component += batou.lib.svn.Checkout(
        'file://' + repos_path, target='clone', revision='head')
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, 'mycomponent/clone/foo'))
    root.component.deploy()  # trigger verify



from batou import CycleErrorDetected, UnknownComponentConfigurationError
from batou import UnsatisfiedResources, UnusedResources
from batou.component import Component, ComponentDefinition
from batou.environment import Environment
import pytest


class Provider(Component):

    def configure(self):
        self.provide('the-answer', 42)


class Consumer(Component):

    def configure(self):
        self.the_answer = self.require('the-answer')


class AggressiveConsumer(Component):

    def configure(self):
        self.the_answer = self.require('the-answer')[0]


class SameHostConsumer(Component):

    def configure(self):
        self.the_answer = self.require('the-answer', self.host)


class Broken(Component):

    def configure(self):
        raise KeyError('foobar')


class Circular1(Component):

    def configure(self):
        self.asdf = self.require('asdf')
        self.provide('bsdf', self)


class Circular2(Component):

    def configure(self):
        self.bsdf = self.require('bsdf')
        self.provide('asdf', self)


class DirtySingularCircularReverse(Component):

    def configure(self):
        self.bsdf = self.require('bsdf', reverse=True, dirty=True)
        self.provide('asdf', self)


@pytest.fixture
def env():
    env = Environment('test')
    for component in globals().values():
        if not isinstance(component, type):
            continue
        if issubclass(component, Component):
            compdef = ComponentDefinition(component)
            env.components[compdef.name] = compdef
    return env


def test_provider_without_consumer_raises_error(env):
    env.add_root('provider', 'host')
    with pytest.raises(UnusedResources):
        env.configure()


def test_consumer_retrieves_value_from_provider_order1(env):
    provider = env.add_root('provider', 'test')
    consumer = env.add_root('consumer', 'test')
    env.configure()
    assert list(consumer.component.the_answer) == [42]
    assert env.roots_in_order() == [provider, consumer]


def test_provider_with_consumer_limited_by_host_raises_error(env):
    env.add_root('provider', 'test2')
    env.add_root('samehostconsumer', 'test')
    with pytest.raises(UnusedResources):
        env.configure()


def test_consumer_retrieves_value_from_provider_order2(env):
    consumer = env.add_root('consumer', 'host')
    provider = env.add_root('provider', 'host')
    env.configure()
    assert list(consumer.component.the_answer) == [42]
    assert env.roots_in_order() == [provider, consumer]


def test_consumer_without_provider_raises_error(env):
    env.add_root('consumer', 'host')
    with pytest.raises(Exception):
        env.configure()
    for exc in env.exceptions:
        if isinstance(exc, UnsatisfiedResources):
            assert set(['the-answer']) == exc.resources
            break
    else:
        assert False, "Did not find exception"


def test_aggressive_consumer_raises_unsatisfiedrequirement(env):
    env.add_root('aggressiveconsumer', 'host')
    with pytest.raises(Exception):
        env.configure()
    for exc in env.exceptions:
        if isinstance(exc, UnsatisfiedResources):
            assert set(['the-answer']) == exc.resources
            break
    else:
        assert False, "Did not find expected exception."


def test_broken_component_logs_real_exception(env):
    env.add_root('broken', 'host')
    with pytest.raises(UnknownComponentConfigurationError):
        env.configure()


def test_consumer_retrieves_value_from_provider_with_same_host(env):
    consumer = env.add_root('samehostconsumer', 'host')
    provider = env.add_root('provider', 'host')
    consumer2 = env.add_root('samehostconsumer', 'host2')
    provider2 = env.add_root('provider', 'host2')
    env.configure()
    assert list(consumer.component.the_answer) == [42]
    assert list(consumer2.component.the_answer) == [42]
    components = env.roots_in_order()
    assert set([provider2, provider]) == set(components[:2])
    assert set([consumer, consumer2]) == set(components[2:])


def test_components_are_ordered_over_multiple_hosts(env):
    provider1 = env.add_root('provider', 'host')
    provider2 = env.add_root('provider', 'host2')
    consumer1 = env.add_root('consumer', 'host')
    consumer2 = env.add_root('consumer', 'host2')
    env.configure()
    components = env.roots_in_order()
    assert set([provider1, provider2]) == set(components[:2])
    assert set([consumer1, consumer2]) == set(components[2:])


def test_circular_depending_component(env):
    env.add_root('circular1', 'test')
    env.add_root('circular2', 'test')
    with pytest.raises(CycleErrorDetected):
        env.configure()


def test_dirty_dependency_for_one_time_retrieval(env, capsys):
    env.add_root('circular1', 'test')
    env.add_root('dirtysingularcircularreverse', 'test')
    env.configure()



from batou import UpdateNeeded, output
from batou.component import Component, Attribute, handle_event
from batou.lib.buildout import Buildout
from batou.lib.file import File, Directory
from batou.lib.nagios import ServiceCheck
from batou.lib.logrotate import RotatedLogfile
from batou.lib.service import Service
from batou.utils import Address
import os
import os.path
import time


class Program(Component):

    program_section = """
[program:{{component.name}}]
command = {{component.command}} {{component.args}}
process_name = {{component.name}}
directory = {{component.directory}}
priority = {{component.priority}}
stderr_logfile = {{component.supervisor.logdir.path}}/{{component.name}}.log
stderr_logfile_backups = 0
stderr_logfile_maxbytes = 0
stdout_logfile = {{component.supervisor.logdir.path}}/{{component.name}}.log
stdout_logfile_backups = 0
stdout_logfile_maxbytes = 0
redirect_stderr = true
{% for k, v in component.options.items() %}
{{k}} = {{v}}
{% endfor -%}
"""

    namevar = 'name'

    # Can this be deployed while running or should the process be stopped
    # prior to deploying it?
    # hot = it can keep running
    # cold = needs to be stopped
    deployment = 'hot'

    command = None
    command_absolute = True
    options = {}
    args = ''
    priority = 10
    directory = None
    dependencies = None

    program_template = ('{priority} {name} ({options}) {command} '
                        '{args} {directory} true')

    def configure(self):
        self.supervisor = self.require_one('supervisor', self.host)
        if self.dependencies is None:
            self.dependencies = (self.parent,)
        if not self.command:
            raise ValueError(
                '`command` option missing for program {}'.format(self.name))
        if not self.directory:
            self.directory = self.workdir
        if self.command_absolute:
            self.command = os.path.normpath(
                os.path.join(self.workdir, self.command))

        if 'startsecs' not in self.options:
            self.options['startsecs'] = 5
        if 'startretries' not in self.options:
            self.options['startretries'] = 3
        self.supervisor.max_startup_delay = max(
            int(self.options['startsecs']), self.supervisor.max_startup_delay)

        self.config = self.expand(
            '{{component.supervisor.program_config_dir.path}}/'
            '{{component.name}}.conf')
        self += File(self.config,
                     content=self.expand(self.program_section))

    def ctl(self, args, **kw):
        command = '{}/bin/supervisorctl'.format(self.supervisor.workdir)
        return self.cmd('{} {}'.format(command, args), **kw)

    def verify(self):
        if not self.supervisor.enable:
            return
        for dependency in self.dependencies:
            dependency.assert_no_changes()
        out, err = self.ctl('status {}'.format(self.name))
        if 'RUNNING' not in out:
            raise UpdateNeeded()

    def update(self):
        self.ctl('reread')
        self.ctl('update')
        self.ctl('restart {}'.format(self.name))
        if self.supervisor.wait_for_running:
            for retry in range(self.options['startsecs']):
                time.sleep(1)
                out, err = self.ctl('status {}'.format(self.name))
                if 'RUNNING' in out:
                    return
            raise RuntimeError(
                'Program "{}" did not start up'.format(self.name))

    # Keep track whether
    _evaded = False

    @handle_event('before-update', 'precursor')
    def evade(self, component):
        if self.deployment == 'hot':
            return
        if self._evaded:
            return
        # Only try once. Keep going anyway.
        self._evaded = True
        output.annotate(
            u"\u2623 Stopping {} for cold deployment".format(self.name))
        try:
            self.ctl('stop {}'.format(self.name))
        except Exception:
            pass


class Eventlistener(Program):

    program_section = """
[eventlistener:{{component.name}}]
command = {{component.command}} {{component.args}}
events = {{component.events}}
process_name={{component.name}}
"""

    namevar = 'name'

    events = ('TICK_60',)
    command = None
    args = None

    def configure(self):
        if not isinstance(self.events, str):
            self.events = ','.join(self.events)
            # Not sure what's right. We only use eventlisteners with superlance
            # which lives in the supervisor's workdir. However, the
            # EventListener component gets instanciated as a sub-component of
            # the actual component using it - which doesn't know about the path
            # to the superlance plugins. :/
            self.supervisor = self.require_one('supervisor', self.host)
            self.command = os.path.normpath(
                os.path.join(self.supervisor.workdir, self.command))

        super(Eventlistener, self).configure()


class Supervisor(Component):

    address = Attribute(Address, 'localhost:9001')
    buildout_cfg = os.path.join(
        os.path.dirname(__file__), 'resources', 'supervisor.buildout.cfg')
    supervisor_conf = os.path.join(
        os.path.dirname(__file__), 'resources', 'supervisor.conf')

    program_config_dir = None
    logdir = None
    loglevel = 'info'

    logrotate = Attribute('literal', 'False')
    nagios = Attribute('literal', 'False')

    # Allows turning "everything off" via environment configuration
    enable = Attribute('literal', 'True')
    # Hot deployments: if supervisor is already running stuff - keep them
    # running
    # Cold deployments: if supervisor is already running: let it run
    # but shutdown all processes before continuing
    deployment_mode = Attribute(str, 'hot')
    max_startup_delay = Attribute(int, 0)
    wait_for_running = Attribute('literal', 'True')
    pidfile = Attribute(str, 'supervisord.pid', map=True)
    socketpath = Attribute(
        str, '{{component.workdir}}/supervisor.sock', map=True)
    check_contact_groups = None

    def configure(self):
        self.provide('supervisor', self)

        buildout_cfg = File('buildout.cfg',
                            source=self.buildout_cfg)
        self += Buildout('buildout',
                         version='2.5.0',
                         setuptools='19.2',
                         config=buildout_cfg,
                         python='2.7')

        self.program_config_dir = Directory('etc/supervisor.d', leading=True)
        self += self.program_config_dir
        self += File('etc/supervisord.conf',
                     source=self.supervisor_conf)
        self.logdir = Directory('var/log', leading=True)
        self += self.logdir

        postrotate = self.expand(
            'kill -USR2 $({{component.workdir}}/bin/supervisorctl pid)')

        if self.logrotate:
            self += RotatedLogfile('var/log/*.log', postrotate=postrotate)

        self += Service('bin/supervisord', pidfile=self.pidfile)
        service = self._

        if self.enable:
            self += RunningSupervisor(service)
        else:
            self += StoppedSupervisor(service)

        # Nagios check
        if self.nagios:
            self += File('check_supervisor',
                         mode=0o755,
                         source=os.path.join(
                             os.path.dirname(__file__),
                             'resources',
                             'check_supervisor.py.in'))

            self += ServiceCheck(
                'Supervisor programs',
                nrpe=True,
                contact_groups=self.check_contact_groups,
                command=self.expand('{{component.workdir}}/check_supervisor'))


class RunningSupervisor(Component):

    action = None
    reload_timeout = 60

    namevar = 'service'

    def verify(self):
        self.parent.assert_no_changes()
        self.assert_file_is_current(
            self.parent.pidfile, ['bin/supervisord', 'etc/supervisord.conf'])
        if not self.is_running():
            raise UpdateNeeded()

    def is_running(self):
        pid, err = self.cmd('bin/supervisorctl pid')
        try:
            int(pid) > 0
        except ValueError:
            return False
        return True

    def update(self):
        if not self.is_running():
            self.service.start()
        else:
            self.reload_supervisor()
        if self.parent.wait_for_running:
            # Wait max startup time now that supervisor is back
            time.sleep(self.parent.max_startup_delay)

    def reload_supervisor(self):
        self.cmd('bin/supervisorctl reload')
        # Reload is asynchronous and doesn't wait for supervisor to become
        # fully running again. We actually could monitor supervisorctl,
        # though. This can take a long time if supervisor needs to orderly
        # shut down a lot of services.
        wait = self.reload_timeout
        while wait:
            out, err = '', ''
            # Supervisor tends to "randomly" set zero and non-zero exit codes.
            # See https://github.com/Supervisor/supervisor/issues/24 :-/
            out, err = self.cmd('bin/supervisorctl pid',
                                ignore_returncode=True)
            if 'SHUTDOWN_STATE' in out:
                time.sleep(1)
                continue
            try:
                int(out) > 0
            except ValueError:
                time.sleep(1)
                wait -= 1
            else:
                break
        else:
            raise RuntimeError(
                'supervisor master process did not start within {} seconds'
                .format(self.reload_timeout))


class StoppedSupervisor(Component):

    def verify(self):
        out, err = self.cmd('bin/supervisorctl pid')
        try:
            int(out)
        except ValueError:
            return
        raise UpdateNeeded()

    def update(self):
        self.cmd('bin/supervisorctl shutdown')



from batou.component import Component


class Service(Component):
    """A generic component to provide a system service.

    Platform-specific components need to perform the work necessary
    to ensure startup and shutdown of the executable correctly.
    """

    namevar = 'executable'

    pidfile = None  # The pidfile as written by the services' executable.

    def start(self):
        """Start service.

        The actual work *should* be done by platform-specific componets. If
        there is no platform-specific component handling the start, the
        executable is executed for reasons of backward compatibility.

        """
        if self._platform_component is not None:
            assert self._platform_component._prepared
            start = getattr(self._platform_component, 'start', None)
            if callable(start):
                start()
                return
        self.cmd(self.executable)



from StringIO import StringIO
from batou.lib.cmmi import Build, Configure, Make
from batou.lib.file import File
from datetime import datetime
import hashlib
import mock
import os.path
import pytest
import shutil
import sys
import tarfile
import time


def test_build_breadcrumb_shortens_url():
    b = Build('http://launchpad.net/libmemcached/1.0/0.46/'
              '+download/libmemcached-0.46.tar.gz')
    assert b._breadcrumb == 'Build(libmemcached-0.46.tar.gz)'


def test_configure_defaults_prefix_to_workdir(root):
    configure = Configure('.')
    configure.cmd = mock.Mock()
    root.component += configure
    assert configure.prefix == root.component.workdir
    root.component.deploy()
    assert (
        ' --prefix={}'.format(root.component.workdir) in
        configure.cmd.call_args[0][0])


def test_configure_accepts_custom_prefix(root):
    configure = Configure('.', prefix='/asdf')
    configure.cmd = mock.Mock()
    root.component += configure
    assert configure.prefix == '/asdf'
    root.component.deploy()
    assert ' --prefix=/asdf' in configure.cmd.call_args[0][0]


def test_configure_verifies_against_success_file(root):
    root.component += File('foo/config.status', content='', leading=True)
    configure = Configure('foo')
    configure.cmd = mock.Mock()
    root.component += configure
    root.component.deploy()
    root.component.deploy()
    assert 1 == configure.cmd.call_count


def test_make_verifies_against_success_file(root):
    root.component += File('foo/Makefile', content='', leading=True)
    make = Make('foo')
    make.cmd = mock.Mock()
    root.component += make
    root.component.deploy()
    root.component.deploy()
    assert 1 == make.cmd.call_count


CONFIGURE_TEMPLATE = """#!%s
import sys
open('configure-running', 'w').close()

Makefile_template = '''
all:
\ttouch make-all

install:
\ttouch make-install
'''

open('Makefile', 'w').write(Makefile_template)
"""


@pytest.fixture
def cmmi_tar(tmpdir):
    # inspired by <http://svn.zope.org/repos/main/zc.recipe.cmmi/trunk
    #              /src/zc/recipe/cmmi/tests.py>
    tarpath = str(tmpdir / 'example-build.tgz')
    tar = tarfile.open(tarpath, 'w:gz')
    now = time.mktime(datetime.now().timetuple())

    info = tarfile.TarInfo('folder')
    info.type = tarfile.DIRTYPE
    info.mtime = now
    tar.addfile(info)

    configure = CONFIGURE_TEMPLATE % sys.executable
    info = tarfile.TarInfo('folder/configure')
    info.size = len(configure)
    info.mode = 0755
    info.mtime = now
    tar.addfile(info, StringIO(configure))

    tar.close()

    fixture = type('Dummy', (object,), {})()
    fixture.path = tarpath
    fixture.checksum = hashlib.md5(open(tarpath, 'r').read()).hexdigest()
    return fixture


@pytest.mark.slow
def test_runs_cmmi(root, cmmi_tar):
    c = Build(cmmi_tar.path, checksum='md5:' + cmmi_tar.checksum)
    root.component += c

    def copy(uri, target):
        shutil.copyfile(uri, target)
        return target, []
    with mock.patch('batou.lib.download.urlretrieve', new=copy):
        root.component.deploy()

    assert os.path.isfile(os.path.join(
        root.environment.workdir_base,
        'mycomponent/example-build/make-install'))



# This should be only one line. If it must be multi-line, indent the second
# line onwards to keep the PKG-INFO file format intact.
"""A utility for automating multi-host, multi-environment software builds \
and deployments.
"""

from setuptools import setup, find_packages
import glob
import os.path


def project_path(*names):
    return os.path.join(*names)


version = open('src/batou/version.txt').read().strip()

setup(
    name='batou',
    version=version,
    install_requires=[
        'Jinja2',
        'requests',
        'setuptools',
        'execnet',
        'py'
    ],
    extras_require={
        'test': [
        ],
    },
    entry_points="""
        [console_scripts]
            batou-init = batou.init:console_main
        [zc.buildout]
            requirements = batou.buildout:Requirements
        [zest.releaser.prereleaser.after]
            update_requirements = batou.release:update_requirements
        [zest.releaser.postreleaser.after]
            update_requirements = batou.release:update_requirements
    """,
    author='Christian Theune',
    author_email='ct@flyingcircus.io',
    license='BSD (2-clause)',
    url='https://plan.flyingcircus.io/projects/batou',
    keywords='deployment',
    classifiers="""\
License :: OSI Approved :: BSD License
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 2 :: Only
"""[:-1].split('\n'),
    description=__doc__.strip(),
    long_description='\n\n'.join(open(project_path(name)).read() for name in (
            'README',
            'CHANGES.txt',
            )),

    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    data_files=[('', glob.glob(project_path('*.txt')))],
    zip_safe=False,
    test_suite='batou.tests',
)
